/// <reference types="node" />
/// <reference types="node" />
import { EventEmitter } from "events";
import * as Protobuf from "protobufjs";
import { ChannelCredentials, Client, ClientOptions } from "@grpc/grpc-js";
import { ProtoRequest, StreamReader, RequestOptions, StreamWriterSandbox } from "./ProtoRequest";
import type { VerifyOptions } from "@grpc/grpc-js/build/src/channel-credentials";
declare type AnyProtoRequest = ProtoRequest<any, any>;
declare type EndpointMatcher = (method: string, request: AnyProtoRequest) => boolean;
/**
 * Middleware to run before each request
 * @param request Initialized, but not started, gRPC request
 * @param client Parent client instance
 */
export declare type RequestMiddleware = (request: AnyProtoRequest, client: ProtoClient) => Promise<void>;
/**
 * Service endpoint configuration
 */
export interface ClientEndpoint {
    /**
     * Remote server address (with port built in)
     */
    address: string;
    /**
     * For configuring secure credentials when connecting to service endpoint
     */
    credentials?: ChannelCredentials | {
        /**
         * Root certificate data
         */
        rootCerts?: Buffer;
        /**
         * Client certificate private key, if available
         */
        privateKey?: Buffer;
        /**
         * Client certificate key chain, if available
         */
        certChain?: Buffer;
        /**
         * Additional options to modify certificate verification
         */
        verifyOptions?: VerifyOptions;
    };
    /**
     * gRPC client options for a connection
     */
    clientOptions?: ClientOptions;
    /**
     * Custom matching of proto method namespace to client endpoint
     */
    match?: string | RegExp | EndpointMatcher;
}
/**
 * Client Settings
 */
export interface ClientSettings {
    /**
     * Either
     */
    endpoint: string | ClientEndpoint | ClientEndpoint[];
    /**
     * Indicates if error should be thrown when caller cancels the request
     */
    rejectOnAbort?: true;
    /**
     * Time in milliseconds before cancelling the request. Defaults to 0 for no timeout
     */
    timeout?: number;
    /**
     * Retry logic for every request
     */
    retryOptions?: RequestOptions["retryOptions"];
}
/**
 * Protobuf parsing settings
 */
export interface ProtoSettings {
    /**
     * Custom root protobuf namespace, for skipping proto file paths
     */
    root?: Protobuf.Root;
    /**
     * Proto file path(s)
     */
    files?: string | string[];
    /**
     * Parsing options for proto files passed
     */
    parseOptions?: Protobuf.IParseOptions;
    /**
     * Message conversion options
     */
    conversionOptions?: Protobuf.IConversionOptions;
}
export interface ProtoClient {
    on(event: "request", listener: (request: AnyProtoRequest) => void): this;
    off(event: "request", listener: (request: AnyProtoRequest) => void): this;
    emit(eventName: "request", request: AnyProtoRequest): boolean;
}
/**
 * Shorthand wrapper utilities for gRPC client calls
 */
export declare class ProtoClient extends EventEmitter {
    /**
     * Client settings
     */
    clientSettings: ClientSettings;
    /**
     * Protobuf parsing settings
     */
    protoSettings: ProtoSettings;
    /**
     * Internal proto conversion options, for referencing
     */
    protoConversionOptions: Protobuf.IConversionOptions;
    /**
     * List of currently active requests
     */
    activeRequests: AnyProtoRequest[];
    /**
     * Middleware to run before each request
     */
    middleware: RequestMiddleware[];
    /**
     * Root protobuf object for referencing
     */
    private root?;
    /**
     * Client endpoints list
     */
    private clients;
    /**
     * Shorthand wrapper utilities for gRPC client calls
     * @param options Configuration options for the client
     */
    constructor(options?: {
        clientSettings?: ClientSettings;
        protoSettings?: ProtoSettings;
        conversionOptions?: Protobuf.IConversionOptions;
    });
    /**
     * Configures settings for client connection
     * @param clientSettings Settings for client connection
     */
    configureClient(clientSettings: ClientSettings): void;
    /**
     * Configures options for reading/writing proto definitions
     * @param protoSettings Proto reading/writing settings
     */
    configureProtos(protoSettings: ProtoSettings): void;
    /**
     * Configures conversion options when decoding proto responses
     * @param options Conversion options
     */
    configureConversionOptions(options: Protobuf.IConversionOptions): void;
    /**
     * Fetches the endpoint matching client instance
     * throwing an error if not yet configured
     * @param request Active request looking for it's client endpoint
     */
    getClient(request: AnyProtoRequest): Client;
    /**
     * Fetches the current proto root object, throwing an error if not yet configured
     */
    getRoot(): Protobuf.Root;
    /**
     * Adds function to the middleware stack
     * @param middleware Middleware to run before each request
     */
    useMiddleware(middleware: RequestMiddleware): void;
    /**
     * Sends unary (request/response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     */
    makeUnaryRequest<ResponseType>(method: string): Promise<ProtoRequest<unknown, ResponseType | undefined>>;
    /**
     * Sends unary (request/response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param data Data to be sent as part of the request
     */
    makeUnaryRequest<RequestType, ResponseType>(method: string, data: RequestType): Promise<ProtoRequest<RequestType, ResponseType | undefined>>;
    /**
     * Sends unary (request/response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param data Data to be sent as part of the request. Defaults to empty object
     * @param abortController Abort controller for canceling the active request
     */
    makeUnaryRequest<RequestType, ResponseType>(method: string, data: RequestType | null, abortController: AbortController): Promise<ProtoRequest<RequestType, ResponseType | undefined>>;
    /**
     * Sends unary (request/response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param data Data to be sent as part of the request. Defaults to empty object
     * @param requestOptions Options for this specific request
     */
    makeUnaryRequest<RequestType, ResponseType>(method: string, data: RequestType | null, requestOptions: RequestOptions): Promise<ProtoRequest<RequestType, ResponseType | undefined>>;
    /**
     * Sends client stream (request stream, single response) request to the service endpoint.
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param writerSandbox Async supported callback for writing data to the open stream
     */
    makeClientStreamRequest<RequestType, ResponseType>(method: string, writerSandbox: StreamWriterSandbox<RequestType, ResponseType>): Promise<ProtoRequest<RequestType, ResponseType | undefined>>;
    /**
     * Sends client stream (request stream, single response) request to the service endpoint.
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param writerSandbox Async supported callback for writing data to the open stream
     * @param abortController Abort controller for canceling the active request
     */
    makeClientStreamRequest<RequestType, ResponseType>(method: string, writerSandbox: StreamWriterSandbox<RequestType, ResponseType>, abortController: AbortController): Promise<ProtoRequest<RequestType, ResponseType | undefined>>;
    /**
     * Sends client stream (request stream, single response) request to the service endpoint.
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param writerSandbox Async supported callback for writing data to the open stream
     * @param requestOptions Options for this specific request
     */
    makeClientStreamRequest<RequestType, ResponseType>(method: string, writerSandbox: StreamWriterSandbox<RequestType, ResponseType>, requestOptions: RequestOptions): Promise<ProtoRequest<RequestType, ResponseType | undefined>>;
    /**
     * Sends a server stream (single request, streamed response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param streamReader Iteration function that will be called on every response chunk
     */
    makeServerStreamRequest<ResponseType>(method: string, streamReader: StreamReader<unknown, ResponseType>): Promise<ProtoRequest<unknown, ResponseType>>;
    /**
     * Sends a server stream (single request, streamed response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param data Data to be sent as part of the request
     * @param streamReader Iteration function that will be called on every response chunk
     */
    makeServerStreamRequest<RequestType, ResponseType>(method: string, data: RequestType, streamReader: StreamReader<RequestType, ResponseType>): Promise<ProtoRequest<RequestType, ResponseType>>;
    /**
     * Sends a server stream (single request, streamed response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param data Data to be sent as part of the request
     * @param streamReader Iteration function that will be called on every response chunk
     * @param abortController Abort controller for canceling the active request
     */
    makeServerStreamRequest<RequestType, ResponseType>(method: string, data: RequestType, streamReader: StreamReader<RequestType, ResponseType>, abortController: AbortController): Promise<ProtoRequest<RequestType, ResponseType>>;
    /**
     * Sends a server stream (single request, streamed response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param data Data to be sent as part of the request
     * @param streamReader Iteration function that will be called on every response chunk
     * @param requestOptions Options for this specific request
     */
    makeServerStreamRequest<RequestType, ResponseType>(method: string, data: RequestType, streamReader: StreamReader<RequestType, ResponseType>, requestOptions: RequestOptions): Promise<ProtoRequest<RequestType, ResponseType>>;
    /**
     * Sends a bi-directional (stream request, stream response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param writerSandbox Async supported callback for writing data to the open stream
     * @param streamReader Iteration function that will be called on every response chunk
     */
    makeBidiStreamRequest<RequestType, ResponseType>(method: string, writerSandbox: StreamWriterSandbox<RequestType, ResponseType>, streamReader: StreamReader<RequestType, ResponseType>): Promise<ProtoRequest<RequestType, ResponseType>>;
    /**
     * Sends a bi-directional (stream request, stream response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param writerSandbox Async supported callback for writing data to the open stream
     * @param streamReader Iteration function that will be called on every response chunk
     * @param abortController Abort controller for canceling the active request
     */
    makeBidiStreamRequest<RequestType, ResponseType>(method: string, writerSandbox: StreamWriterSandbox<RequestType, ResponseType>, streamReader: StreamReader<RequestType, ResponseType>, abortController: AbortController): Promise<ProtoRequest<RequestType, ResponseType>>;
    /**
     * Sends a bi-directional (stream request, stream response) request to the service endpoint
     * @param method Fully qualified path of the method for the request that can be used by protobufjs.lookup
     * @param writerSandbox Async supported callback for writing data to the open stream
     * @param streamReader Iteration function that will be called on every response chunk
     * @param requestOptions Options for this specific request
     */
    makeBidiStreamRequest<RequestType, ResponseType>(method: string, writerSandbox: StreamWriterSandbox<RequestType, ResponseType>, streamReader: StreamReader<RequestType, ResponseType>, requestOptions: RequestOptions): Promise<ProtoRequest<RequestType, ResponseType>>;
    /**
     * Aborts all requests that match the method defined
     * @param match Method to match
     */
    abortRequests(match: string): void;
    /**
     * Aborts all requests whose method passes the regex test
     * @param match Regex run against the method namespace for matching
     */
    abortRequests(match: RegExp): void;
    /**
     * Aborts all requests that pass the match function
     * @param match Endpoint matcher util
     */
    abortRequests(match: EndpointMatcher): void;
    /**
     * Aborts all active requests and closes all gRPC connections
     */
    close(): void;
    /**
     * Manages request within the active requests list
     * @param method Method name for the request
     * @param requestMethodType Method type
     */
    private generateRequest;
    /**
     * Removes request from the active request list
     * @param request Request to be removed
     */
    private removeActiveRequest;
    /**
     * Generates a function for matching namespaces to requests
     * @param match Multi-type endpoint matcher to convert
     * @returns Matcher function
     */
    private generateEndpointMatcher;
}
export {};
//# sourceMappingURL=ProtoClient.d.ts.map