/// <reference types="node" />
import * as Protobuf from "protobufjs";
import { EventEmitter } from "events";
import { CallOptions, Metadata, MetadataValue, ServiceError, status, StatusObject } from "@grpc/grpc-js";
import type { ProtoClient } from "./ProtoClient";
import { RequestMethodType } from "./constants";
/**
 * Iteration callback for streamed responses
 * @param row Streamed response row
 * @param rowIndex Index for the current chunked row
 * @param request Currently active request
 */
export declare type StreamReader<RequestType, ResponseType> = (row: ResponseType, rowIndex: number, request: ProtoRequest<RequestType, ResponseType>) => Promise<void>;
/**
 * Writing wrapper for streamed requests
 * @param write Async function for sending an object over the write stream
 * @param request Currently active request
 */
export declare type StreamWriterSandbox<RequestType, ResponseType> = (write: StreamWriter<RequestType>, request: ProtoRequest<RequestType, ResponseType>) => Promise<void>;
/**
 * Writing function for sending objects to the write stream
 * @param data Request data row to be streamed
 * @param encoding Write encoding for the data
 */
export declare type StreamWriter<RequestType> = (data: RequestType, encoding?: string) => Promise<void>;
/**
 * Custom retry logic options
 */
export interface RequestRetryOptions {
    /**
     * Number of times to retry request. Defaults to none
     */
    retryCount?: number;
    /**
     * Status codes request is allowed to retry on
     */
    status?: status | status[];
    /**
     * Indicates if retry should occur after internal client timeout. Defaults to true
     */
    retryOnClientTimeout?: boolean;
}
/**
 * Request specific options
 */
export interface RequestOptions {
    /**
     * Controller for aborting the active request
     */
    abortController?: AbortController;
    /**
     * Metadata to be attached to the request
     */
    metadata?: Record<string, MetadataValue | MetadataValue[]> | Metadata;
    /**
     * Request specific options
     */
    callOptions?: CallOptions;
    /**
     * Time in milliseconds before cancelling the request. Defaults to 0 for no timeout
     */
    timeout?: number;
    /**
     * Indicates retry logic that should be applied to the request
     * @alias boolean Retries the request once for default status code failures when true, disable retry when false
     * @alias number Number of retries to allow for default status code failures
     * @alias RequestRetryOptions Custom retry options
     */
    retryOptions?: boolean | number | RequestRetryOptions;
}
/**
 * Custom event typings
 */
export interface ProtoRequest<RequestType, ResponseType> {
    on(event: "response", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    on(event: "retry", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    on(event: "aborted", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    on(event: "end", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    off(event: "response", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    off(event: "retry", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    off(event: "aborted", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    off(event: "end", listener: (request: ProtoRequest<RequestType, ResponseType>) => void): this;
    emit(eventName: "response", request: ProtoRequest<RequestType, ResponseType>): boolean;
    emit(eventName: "retry", request: ProtoRequest<RequestType, ResponseType>): boolean;
    emit(eventName: "aborted", request: ProtoRequest<RequestType, ResponseType>): boolean;
    emit(eventName: "end", request: ProtoRequest<RequestType, ResponseType>): boolean;
}
/**
 * Individual gRPC request
 */
export declare class ProtoRequest<RequestType, ResponseType> extends EventEmitter {
    /**
     * Fully qualified path of the method for the request that can be used by protobufjs.lookup
     */
    readonly method: string;
    /**
     * Generated request path
     */
    readonly requestPath: string;
    /**
     * Request proto message type
     */
    readonly requestType: Protobuf.Type;
    /**
     * Response proto message type
     */
    readonly responseType: Protobuf.Type;
    /**
     * Request method type
     */
    readonly requestMethodType: RequestMethodType;
    /**
     * Time in milliseconds before cancelling the request
     */
    timeout: number;
    /**
     * Configured retry options for this request
     */
    retryOptions: RequestRetryOptions;
    /**
     * AbortController tied to the request
     */
    abortController: AbortController;
    /**
     * Metadata instance for the request
     */
    metadata: Metadata;
    /**
     * Request specific options
     */
    callOptions: CallOptions;
    /**
     * Data response from the service, only valid for unary and client stream requests
     */
    result?: ResponseType;
    /**
     * Metadata returned from the service
     */
    responseMetadata?: Metadata;
    /**
     * Metadata returned from the service
     */
    responseStatus?: StatusObject;
    /**
     * Number of retries made for this request
     */
    retries: number;
    /**
     * References any error that may have occured during the request
     */
    error?: Error;
    /**
     * When retries are enabled, all errors will be stored here
     */
    readonly responseErrors: Error[];
    /**
     * Internal reference to the parent client instance
     */
    private readonly client;
    /**
     * Internal reference to determine if request has start yet or not
     */
    private hasRequestBegun;
    /**
     * Internal class for managing individual requests
     */
    constructor(client: ProtoClient, method: string, requestMethodType: RequestMethodType, requestOptions: AbortController | RequestOptions | undefined);
    /**
     * Sends unary (request/response) request to the service endpoint
     * @param data Data to be sent as part of the request. Defaults to empty object
     */
    makeUnaryRequest(data?: RequestType | null): Promise<this>;
    /**
     * Sends client stream (request stream, single response) request to the service endpoint
     * @param writerSandbox Async supported callback for writing data to the open stream
     */
    makeClientStreamRequest(writerSandbox: StreamWriterSandbox<RequestType, ResponseType>): Promise<this>;
    /**
     * Sends a server stream (single request, streamed response) request to the service endpoint
     * @param data Data to be sent as part of the request
     * @param streamReader Iteration function that will get called on every response chunk
     */
    makeServerStreamRequest(data: RequestType | StreamReader<RequestType, ResponseType>, streamReader?: StreamReader<RequestType, ResponseType>): Promise<this>;
    /**
     * Sends a bi-directional (stream request, stream response) request to the service endpoint
     * @param writerSandbox Async supported callback for writing data to the open stream
     * @param streamReader Iteration function that will get called on every response chunk
     */
    makeBidiStreamRequest(writerSandbox: StreamWriterSandbox<RequestType, ResponseType>, streamReader: StreamReader<RequestType, ResponseType>): Promise<this>;
    /**
     * Aborts the request if it is still active
     */
    abort(): void;
    /**
     * Serializing method for outgoing messages
     * @param object Request object passed from the caller
     */
    private serializeRequest;
    /**
     * Deserializing method for incoming messages
     * @param buffer Buffer object repsonse from the connection
     */
    private deserializeResponse;
    /**
     * Determins if this request can be retried on error
     * @param code Status code of the current error
     */
    private canRetry;
    /**
     * Wraps active request calls with retry and timeout handlers
     * @param requestRunner Request wrapper for running through retries
     */
    retryWrapper(requestRunner: (
    /**
     * Starts the timer for timeouts, and triggers the reached callback
     * once timer threshold has been met
     */
    timeout: (timeoutReached: () => void) => void, 
    /**
     * Callback for when the service response has completed
     */
    completed: (error?: Error | ServiceError | null) => void) => void): Promise<this>;
}
//# sourceMappingURL=ProtoRequest.d.ts.map